from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.utils import InterfaceError


class DatabaseFeatures(BaseDatabaseFeatures):
    has_bulk_insert = False
    can_use_chunked_reads = True
    supports_microsecond_precision = False
    supports_regex_backreferencing = False
    supports_subqueries_in_group_by = False
    supports_transactions = True
    can_rollback_ddl = True
    uses_savepoints = True
    can_release_savepoints = True
    autocommits_when_autocommit_is_off = False
    supports_timezones = False
    supports_sequence_reset = False
    supports_tablespaces = True
    can_introspect_autofield = True
    has_select_for_update = True
    supports_select_for_update_with_limit = True
    closed_cursor_error_class = InterfaceError



from django.db import models
from django.core.exceptions import ValidationError


class TrimCharField(models.CharField):
    description = "CharField that ignores trailing spaces in data"

    def from_db_value(self, value, expression, connection, *ignore):
        if value:
            return value.rstrip()
        return value


class CharToBooleanField(models.CharField):
    def __init__(self, *args, null=True, **kwargs):
        kwargs['max_length'] = 1

        super().__init__(*args, null=null, **kwargs)

    def from_db_value(self, value, expression, connection, *ignore):
        if self.null and value is None:
            return None

        return value == 'Y'

    def get_db_prep_value(self, value, connection, prepared=False):
        if value is True:
            return 'Y'
        elif value is False:
            return 'N'

        elif value is None and self.null:
            return None

        # - Not sure if this is the right place/thing to do here
        raise ValidationError(
            self.error_messages['invalid_nullable' if self.null else 'null'],
            code='invalid',
            params={'value': value},
        )

    def to_python(self, value):
        return value

    def get_prep_value(self, value):
        return value



EXCLUDED_TABLES = [
    'sysaggregates',
    'sysams',
    'sysattrtypes',
    'sysautolocate',
    'sysblobs',
    'syscasts',
    'syschecks',
    'syscolattribs',
    'syscolauth',
    'syscoldepend',
    'syscolumns',
    'sysconstraints',
    'sysdefaults',
    'sysdepend',
    'sysdirectives',
    'sysdistrib',
    'syserrors',
    'sysextcols',
    'sysextdfiles',
    'sysexternal',
    'sysfragauth',
    'sysfragdist',
    'sysfragments',
    'sysindices',
    'sysinherits',
    'syslangauth',
    'syslogmap',
    'sysobjstate',
    'sysopclasses',
    'sysopclstr',
    'sysprocauth',
    'sysprocbody',
    'sysproccolumns',
    'sysprocedures',
    'sysprocplan',
    'sysreferences',
    'sysroleauth',
    'sysroutinelangs',
    'sysseclabelauth',
    'sysseclabelcomponentelements',
    'sysseclabelcomponents',
    'sysseclabelnames',
    'sysseclabels',
    'syssecpolicies',
    'syssecpolicycomponents',
    'syssecpolicyexemptions',
    'syssequences',
    'syssurrogateauth',
    'syssynonyms',
    'syssyntable',
    'systabamdata',
    'systabauth',
    'systables',
    'systraceclasses',
    'systracemsgs',
    'systrigbody',
    'systriggers',
    'sysusers',
    'sysviews',
    'sysviolations',
    'sysxadatasources',
    'sysxasourcetypes',
    'sysxtddesc',
    'sysxtdtypeauth',
    'sysxtdtypes'
]



"""
informix database backend for Django.

Requires informixdb
"""
import logging
import os
import sys
import platform
import time
import random
import re

from django.db import connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.validation import BaseDatabaseValidation
from django.core.exceptions import ImproperlyConfigured
from django.core import signals
from django.utils.six import binary_type, text_type
from django.utils.encoding import smart_str

from .client import DatabaseClient
from .creation import DatabaseCreation
from .introspection import DatabaseIntrospection
from .operations import DatabaseOperations
from .features import DatabaseFeatures
from .schema import DatabaseSchemaEditor

try:
    import pyodbc
except ImportError as e:
    e = sys.exc_info()[1]
    raise ImproperlyConfigured("Error loading pyodbc module:{}".format(e))


logger = logging.getLogger(__name__)


def decoder(value, encodings=('utf-8',)):
    """This decoder tries multiple encodings before giving up"""

    if not isinstance(value, binary_type):
        raise ValueError(f"Not a binary type: {value} {type(value)}")

    for enc in encodings:
        try:
            return value.decode(enc)
        except UnicodeDecodeError:
            pass

    raise UnicodeDecodeError("unable to decode `{value}`")


class DatabaseWrapper(BaseDatabaseWrapper):
    vendor = 'informixdb'
    Database = pyodbc

    DRIVER_MAP = {
        'DARWIN': '/Applications/IBM/informix/lib/cli/iclit09b.dylib',
        'LINUX': '/opt/IBM/informix/lib/cli/iclit09b.so',
        'WINDOWS32bit': 'IBM INFORMIX ODBC DRIVER (32-bit)',
        'WINDOWS64bit': 'IBM INFORMIX ODBC DRIVER (64-bit)',
    }

    ISOLATION_LEVEL = {
        'READ_COMMITED': pyodbc.SQL_TXN_READ_COMMITTED,
        'READ_UNCOMMITTED': pyodbc.SQL_TXN_READ_UNCOMMITTED,  # Dirty Read
        'REPEATABLE_READ': pyodbc.SQL_TXN_REPEATABLE_READ,
        'SERIALIZABLE': pyodbc.SQL_TXN_SERIALIZABLE,
    }

    data_types = {
        'AutoField': 'serial',
        'BigAutoField': 'bigserial',
        'BinaryField': 'blob',
        'BooleanField': 'boolean',
        'CharField': 'lvarchar(%(max_length)s)',
        'CommaSeparatedIntegerField': 'lvarchar(%(max_length)s)',
        'DateField': 'date',
        'DateTimeField': 'datetime year to fraction(5)',
        'DecimalField': 'decimal',
        'DurationField': 'interval',
        'FileField': 'lvarchar(%(max_length)s)',
        'FilePathField': 'lvarchar(%(max_length)s)',
        'FloatField': 'smallfloat',
        'IntegerField': 'integer',
        'BigIntegerField': 'bigint',
        'IPAddressField': 'char(15)',
        'GenericIPAddressField': 'char(39)',
        'NullBooleanField': 'boolean',
        'OneToOneField': 'integer',
        'PositiveIntegerField': 'integer',
        'PositiveSmallIntegerField': 'smallint',
        'SlugField': 'lvarchar(%(max_length)s)',
        'SmallIntegerField': 'smallint',
        'TextField': 'lvarchar(%(max_length)s)',
        'TimeField': 'datetime hour to second',
        'UUIDField': 'char(32)',
    }

    data_type_check_constraints = {
        'PositiveIntegerField': '%(column)s >= 0',
        'PositiveSmallIntegerField': '%(column)s >= 0',
    }

    operators = {
        'exact': '= %s',
        'iexact': "= LOWER(%s)",
        'contains': "LIKE %s ESCAPE '\\'",
        'icontains': "LIKE LOWER(%s) ESCAPE '\\'",
        'gt': '> %s',
        'gte': '>= %s',
        'lt': '< %s',
        'lte': '<= %s',
        'startswith': "LIKE %s ESCAPE '\\'",
        'endswith': "LIKE %s ESCAPE '\\'",
        'istartswith': "LIKE LOWER(%s) ESCAPE '\\'",
        'iendswith': "LIKE LOWER(%s) ESCAPE '\\'",
        'regex': 'LIKE %s',
        'iregex': 'LIKE %s',
    }

    # The patterns below are used to generate SQL pattern lookup clauses when
    # the right-hand side of the lookup isn't a raw string (it might be an expression
    # or the result of a bilateral transformation).
    # In those cases, special characters for LIKE operators (e.g. \, *, _) should be
    # escaped on database side.
    #
    # Note: we use str.format() here for readability as '%' is used as a wildcard for
    # the LIKE operator.
    pattern_esc = r"REPLACE(REPLACE(REPLACE({}, '\', '\\'), '%%', '\%%'), '_', '\_')"
    pattern_ops = {
        'contains': "LIKE '%%' ESCAPE '\\' || {} || '%%'",
        'icontains': "LIKE '%%' ESCAPE '\\' || UPPER({}) || '%%'",
        'startswith': "LIKE {} ESCAPE '\\' || '%%'",
        'istartswith': "LIKE UPPER({}) ESCAPE '\\' || '%%'",
        'endswith': "LIKE '%%' ESCAPE '\\' || {}",
        'iendswith': "LIKE '%%' ESCAPE '\\' || UPPER({})",
    }
    client_class = DatabaseClient
    creation_class = DatabaseCreation
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    ops_class = DatabaseOperations
    SchemaEditorClass = DatabaseSchemaEditor
    validation_class = BaseDatabaseValidation

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        options = self.settings_dict.get('OPTIONS', {})

        self._validation_enabled = options.get("VALIDATE_CONNECTION", False)
        self._validation_interval = options.get("VALIDATION_INTERVAL", 300)
        self._next_validation = time.time() + self._validation_interval
        self._validation_query = options.get("VALIDATION_QUERY", "SELECT 1 FROM sysmaster:sysdual")
        self.encodings = options.get('encodings', ('utf-8', 'cp1252', 'iso-8859-1'))
        # make lookup operators to be collation-sensitive if needed
        self.collation = options.get('collation', None)
        if self.collation:
            self.operators = dict(self.__class__.operators)
            ops = {}
            for op in self.operators:
                sql = self.operators[op]
                if sql.startswith('LIKE '):
                    ops[op] = '%s COLLATE %s' % (sql, self.collation)
            self.operators.update(ops)

        self.features = self.features_class(self)
        self.ops = self.ops_class(self)
        self.client = self.client_class(self)
        self.creation = self.creation_class(self)
        self.introspection = self.introspection_class(self)
        self.validation = self.validation_class(self)

    def validate_connection(self):
        """
        This method is invoked at the start of a request to verify an existing
        connection is still functional. This is achieved by doing a simple query
        against the database.
        """
        if not self._validation_enabled or time.time() < self._next_validation:
            return

        self._next_validation = time.time() + self._validation_interval

        # We call close_if_unusable_or_obsolete to ensure obsolete connections
        # are closed before we consider validating them. This will result in
        # close_if_unusable_or_obsolete being called twice since it is also
        # called automatically by django. This is ok since the second call is
        # essentially a no-op.
        self.close_if_unusable_or_obsolete()
        if self.connection is not None and not self.is_usable():
            self.close()

    def get_driver_path(self):
        system = platform.system().upper()
        if system == 'WINDOWS':
            system = system + platform.architecture()[0]
        try:
            return self.DRIVER_MAP[system]
        except KeyError:
            raise ImproperlyConfigured('cannot locate informix driver, please specify')

    def get_connection_params(self):
        settings = self.settings_dict

        if 'DSN' not in settings:
            for k in ['NAME', 'SERVER', 'USER', 'PASSWORD']:
                if k not in settings:
                    raise ImproperlyConfigured('{} is a required setting for an informix connection'.format(k))
        conn_params = settings.copy()

        # Ensure the driver is set in the options
        options = conn_params.get('OPTIONS', {})
        if 'DRIVER' not in options or options['DRIVER'] is None:
            options['DRIVER'] = self.get_driver_path()
        if platform.system().upper() != 'WINDOWS':
            sqlhosts = os.environ.get('INFORMIXSQLHOSTS')
            if not sqlhosts or not os.path.exists(sqlhosts):
                raise ImproperlyConfigured('Cannot find Informix sqlhosts at {}'.format(sqlhosts))
            if not os.path.exists(options['DRIVER']):
                raise ImproperlyConfigured('cannot find Informix driver at {}'.format(options['DRIVER']))
        conn_params['OPTIONS'] = options

        conn_params['AUTOCOMMIT'] = False if 'AUTOCOMMIT' not in conn_params else conn_params['AUTOCOMMIT']

        return conn_params

    def get_new_connection(self, conn_params):
        parts = [
            'Driver={{{}}}'.format(conn_params['OPTIONS']['DRIVER']),
        ]

        if 'DSN' in conn_params:
            parts.append('DSN={}'.format(conn_params['DSN']))
        if 'SERVER' in conn_params:
            parts.append('Server={}'.format(conn_params['SERVER']))
        if 'NAME' in conn_params and conn_params['NAME'] is not None:
            parts.append('Database={}'.format(conn_params['NAME']))
        elif conn_params['NAME'] is None:
            parts.append('CONNECTDATABASE=no')
        if 'USER' in conn_params:
            parts.append('Uid={}'.format(conn_params['USER']))
        if 'PASSWORD' in conn_params:
            parts.append('Pwd={}'.format(conn_params['PASSWORD']))
        if 'CPTIMEOUT' in conn_params['OPTIONS']:
            parts.append('CPTimeout={}'.format(conn_params['OPTIONS']['CPTIMEOUT']))

        connection_string = ';'.join(parts)
        logging.debug('Connecting to Informix')
        self.connection = self._get_connection_with_retries(connection_string, conn_params)
        self.connection.setencoding(encoding='UTF-8')

        # This will set database isolation level at connection level
        if 'ISOLATION_LEVEL' in conn_params['OPTIONS']:
            self.connection.set_attr(pyodbc.SQL_ATTR_TXN_ISOLATION,
                                     self.ISOLATION_LEVEL[conn_params['OPTIONS']['ISOLATION_LEVEL']])

        # This will set SQL_C_CHAR, SQL_C_WCHAR and SQL_BINARY to 32000
        # this max length is actually just what the database internally
        # supports. e.g. the biggest `LONGVARCHAR` field in informix is
        # 32000, you would need to split anything bigger over multiple fields
        # This limit will not effect schema defined lengths, which will just
        # truncate values greater than the limit.
        self.connection.maxwrite = 32000

        self.connection.add_output_converter(-101, lambda r: r.decode('utf-8'))  # Constraints
        self.connection.add_output_converter(-391, lambda r: r.decode('utf-16-be'))  # Integrity Error

        self.connection.add_output_converter(pyodbc.SQL_CHAR, self._output_converter)
        self.connection.add_output_converter(pyodbc.SQL_WCHAR, self._output_converter)
        self.connection.add_output_converter(pyodbc.SQL_VARCHAR, self._output_converter)
        self.connection.add_output_converter(pyodbc.SQL_WVARCHAR, self._output_converter)
        self.connection.add_output_converter(pyodbc.SQL_LONGVARCHAR, self._output_converter)
        self.connection.add_output_converter(pyodbc.SQL_WLONGVARCHAR, self._output_converter)

        if 'LOCK_MODE_WAIT' in conn_params['OPTIONS']:
            self.set_lock_mode(wait=conn_params['OPTIONS']['LOCK_MODE_WAIT'])

        return self.connection

    def _get_connection_with_retries(self, connection_string, conn_params):
        """
        Attempt to open a connection, retrying on failure with an exponential backoff.
        """
        retry_params = conn_params.get("CONNECTION_RETRY", {})
        max_attempts = retry_params.get("MAX_ATTEMPTS", 1)
        wait_min = retry_params.get("WAIT_MIN", 0)
        wait_max = retry_params.get("WAIT_MAX", 1000)
        multiplier = retry_params.get("WAIT_MULTIPLIER", 25)
        exp_base = retry_params.get("WAIT_EXP_BASE", 2)
        errors_to_retry = retry_params.get("ERRORS", ["-908", "-930", "-27001"])
        retryable = re.compile(r"\((" + "|".join(errors_to_retry) + r")\)")

        attempt = 0
        while True:
            attempt += 1
            try:
                conn = pyodbc.connect(connection_string, autocommit=conn_params["AUTOCOMMIT"],
                                      timeout=conn_params["OPTIONS"].get("CONN_TIMEOUT", 0))
            except pyodbc.Error as err:
                if attempt < max_attempts and retryable.search(err.args[1]):
                    wait = random.uniform(
                        wait_min,
                        max(wait_min, min(wait_max, multiplier * exp_base ** (attempt - 1))),
                    )
                    logger.info(
                        f'failed to connect to db on attempt {attempt}: "{err}"; '
                        f"waiting {wait:.1f} ms before trying again"
                    )
                    time.sleep(wait / 1000)
                    continue
                else:
                    raise
            else:
                return conn

    def _unescape(self, raw):
        """
        For some reason the Informix ODBC driver seems to double escape new line characters.

        This little handler converts them back.

        @todo: See if this applies to other escape characters
        """
        return raw.replace(b'\\n', b'\n')

    def _output_converter(self, raw):
        return decoder(self._unescape(raw), self.encodings)

    def init_connection_state(self):
        pass

    def create_cursor(self, name=None):
        logging.debug('Creating Informix cursor')
        return CursorWrapper(self.connection.cursor(), self)

    def _set_autocommit(self, autocommit):
        with self.wrap_database_errors:
            self.connection.autocommit = autocommit

    def check_constraints(self, table_names=None):
        """
        To check constraints, we set constraints to immediate. Then, when, we're done we must ensure they
        are returned to deferred.
        """
        self.cursor().execute('SET CONSTRAINTS ALL IMMEDIATE')
        self.cursor().execute('SET CONSTRAINTS ALL DEFERRED')

    def _start_transaction_under_autocommit(self):
        """
        Start a transaction explicitly in autocommit mode.
        """
        start_sql = self.ops.start_transaction_sql()
        self.cursor().execute(start_sql)

    def is_usable(self):
        # We create a cursor and then explicitly close it as there is a bug
        # that is encountered when relying on garbage collection to close the
        # cursor: https://github.com/mkleehammer/pyodbc/issues/585
        try:
            cursor = self.connection.cursor()
        except pyodbc.Error as exc:
            logger.info(f"error creating cursor: {exc}")
            return False

        try:
            cursor.execute(self._validation_query)
            return True
        except pyodbc.Error as exc:
            logger.info(f"error executing query: {exc}")
            return False
        finally:
            # We close the cursor explicitly to work around the pyodbc bug
            # described at the top of this function. If closing the cursor
            # fails we set the return value to `False`. Otherwise it
            # remains what was returned in the `try` or `except` block, which
            # depends on whether `cursor.execute` succeeded or not.
            try:
                cursor.close()
            except pyodbc.Error as exc:
                logger.info(f"error closing cursor: {exc}")
                return False

    def read_dirty(self):
        self.cursor().execute('set isolation to dirty read;')

    def read_committed(self):
        self.cursor().execute('set isolation to committed read;')

    def read_repeatable(self):
        self.cursor().execute('set isolation to repeatable read;')

    def read_committed_with_update_locks(self):
        self.cursor().execute('set isolation to committed read retain update locks;')

    def set_lock_mode(self, wait=None):
        """
        This will set database LOCK MODE WAIT at connection level
        Application can use this property to override the default server
        process for accessing a locked row or table.
        The default value is 0 (do not wait for the lock).
        Possible values:
           -1 - WAIT until the lock is released.
           0 - DO NOT WAIT, end the operation, and return with error.
           nn - WAIT for nn seconds for the lock to be released.
        """
        if wait == 0:
            sql = 'SET LOCK MODE TO NOT WAIT'
        elif wait == -1:
            sql = 'SET LOCK MODE TO WAIT'
        else:
            sql = 'SET LOCK MODE TO WAIT {}'.format(wait)

        self.cursor().execute(sql)

    def _commit(self):
        if self.connection is not None:
            with self.wrap_database_errors:
                return self.cursor().execute("COMMIT WORK")

    def _rollback(self):
        if self.connection is not None:
            with self.wrap_database_errors:
                return self.cursor().execute("ROLLBACK WORK")


class CursorWrapper(object):
    """
    A wrapper around the pyodbc's cursor that takes in account a) some pyodbc
    DB-API 2.0 implementation and b) some common ODBC driver particularities.
    """
    def __init__(self, cursor, connection):
        self.active = True
        self.cursor = cursor
        self.connection = connection
        self.driver_charset = False  # connection.driver_charset
        self.last_sql = ''
        self.last_params = ()

    def close(self):
        if self.active:
            self.active = False
            self.cursor.close()

    def format_sql(self, sql, params):
        if isinstance(sql, text_type):
            # FreeTDS (and other ODBC drivers?) doesn't support Unicode
            # yet, so we need to encode the SQL clause itself in utf-8
            sql = smart_str(sql, self.driver_charset)

        # pyodbc uses '?' instead of '%s' as parameter placeholder.
        if params is not None:
            pass
            # sql = sql % tuple('?' * len(params))

        return sql

    def format_params(self, params):
        fp = []
        if params is not None:
            for p in params:
                if isinstance(p, text_type):
                    if self.driver_charset:
                        # FreeTDS (and other ODBC drivers?) doesn't support Unicode
                        # yet, so we need to encode parameters in utf-8
                        fp.append(smart_str(p, self.driver_charset))
                    else:
                        fp.append(p)

                elif isinstance(p, binary_type):
                    fp.append(p)

                elif isinstance(p, type(True)):
                    if p:
                        fp.append(1)
                    else:
                        fp.append(0)

                else:
                    fp.append(p)

        return tuple(fp)

    def execute(self, sql, params=None):
        self.last_sql = sql
        sql = self.format_sql(sql, params)
        params = self.format_params(params)
        self.last_params = params
        return self.cursor.execute(sql, params)

    def executemany(self, sql, params_list=()):
        if not params_list:
            return None
        raw_pll = [p for p in params_list]
        sql = self.format_sql(sql, raw_pll[0])
        params_list = [self.format_params(p) for p in raw_pll]
        return self.cursor.executemany(sql, params_list)

    def format_rows(self, rows):
        return list(map(self.format_row, rows))

    def format_row(self, row):
        """
        Decode data coming from the database if needed and convert rows to tuples
        (pyodbc Rows are not sliceable).
        """
        if self.driver_charset:
            for i in range(len(row)):
                f = row[i]
                # FreeTDS (and other ODBC drivers?) doesn't support Unicode
                # yet, so we need to decode utf-8 data coming from the DB
                if isinstance(f, binary_type):
                    row[i] = f.decode(self.driver_charset)
        return tuple(row)

    def fetchone(self):
        row = self.cursor.fetchone()
        if row is not None:
            row = self.format_row(row)
        # Any remaining rows in the current set must be discarded
        # before changing autocommit mode when you use FreeTDS
        self.cursor.nextset()
        return row

    def fetchmany(self, chunk):
        return self.format_rows(self.cursor.fetchmany(chunk))

    def fetchall(self):
        return self.format_rows(self.cursor.fetchall())

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        return getattr(self.cursor, attr)

    def __iter__(self):
        return iter(self.cursor)


def _validate_connection(**kwargs):
    for conn in connections.all():
        if isinstance(conn, DatabaseWrapper):
            conn.validate_connection()


signals.request_started.connect(_validate_connection)



from decimal import Decimal
from collections import namedtuple

import pytest
from django.test import TestCase
from django.db import connection, transaction
from django.core.exceptions import ValidationError

from .models import Donut


RawDonutResult = namedtuple('RawDonut', [
    'id', 'name', 'trim_name', 'is_frosted', 'is_fresh', 'is_only_fresh', 'cost'
])


class DataTypesTestCase(TestCase):
    def test_boolean_false(self):
        d = Donut(name='Apple Fritter')
        self.assertFalse(d.is_frosted)
        d.save()
        d2 = Donut.objects.get(name='Apple Fritter')
        self.assertFalse(d2.is_frosted)

    def test_boolean_true(self):
        d = Donut(name='Apple Fritter', is_frosted=True)
        self.assertTrue(d.is_frosted)
        d.save()
        d2 = Donut.objects.get(name='Apple Fritter')
        self.assertTrue(d2.is_frosted)

    def test_decimal(self):
        d = Donut(name='Apple Fritter', cost=1)
        self.assertEqual(d.cost, 1)
        d.save()
        d2 = Donut.objects.get(name='Apple Fritter')
        self.assertEqual(d2.cost, Decimal(1))

    def test_decimal_precision(self):
        d = Donut(name='Apple Fritter', cost=1.23)
        self.assertEqual(d.cost, 1.23)
        d.save()
        d2 = Donut.objects.get(name='Apple Fritter')
        self.assertEqual(d2.cost, Decimal('1.23'))

    def test_characters(self):
        d = Donut(name='Apple Fritter ')
        self.assertEqual(d.name, 'Apple Fritter ')
        d.save()
        d2 = Donut.objects.get(name='Apple Fritter ')
        self.assertEqual(d2.name, 'Apple Fritter ')

    def test_trim_characters(self):
        d = Donut(trim_name='abc ')
        self.assertEqual(d.trim_name, 'abc ')
        d.save()
        d2 = Donut.objects.get(trim_name='abc')
        self.assertEqual(d2.trim_name, 'abc')

    def test_char_to_boolean_True(self):
        d = Donut(is_fresh=True)
        self.assertEqual(d.is_fresh, True)
        d.save()
        d2 = Donut.objects.get(is_fresh=True)
        self.assertEqual(d2.is_fresh, True)

    def test_char_to_boolean_False(self):
        d = Donut(is_fresh=False)
        self.assertEqual(d.is_fresh, False)
        d.save()
        d2 = Donut.objects.get(is_fresh=False)
        self.assertEqual(d2.is_fresh, False)

    def test_char_to_boolean_None(self):
        d = Donut(is_fresh=None)
        self.assertEqual(d.is_fresh, None)
        d.save()
        d2 = Donut.objects.get(is_fresh=None)
        self.assertEqual(d2.is_fresh, None)

    def test_char_to_boolean_default_value(self):
        d = Donut()
        self.assertEqual(d.is_fresh, False)
        d.save()
        d2 = Donut.objects.get(is_fresh=False)
        self.assertEqual(d2.is_fresh, False)

    def test_char_to_boolean_not_null(self):
        d = Donut(is_only_fresh=None)
        self.assertRaises(ValidationError, d.save)


@pytest.mark.django_db(transaction=True)
class TestDataTypesCharToBooleanRawSQL():
    def _get_raw_donut(self):
        raw_donut = None

        with transaction.atomic():
            with connection.cursor() as cursor:
                cursor.execute(f'select * from datatypes_donut')
                donut = cursor.fetchone()
                raw_donut = RawDonutResult(*donut)

        return raw_donut

    def test_char_to_boolean_check_empty_db_value(self):
        d = Donut()
        d.save()
        raw_donut = self._get_raw_donut()

        assert raw_donut.is_fresh == 'N'
        assert raw_donut.is_only_fresh == 'N'

    @pytest.mark.parametrize('model_value, raw_value', [
        (True, 'Y'),
        (False, 'N'),
    ])
    def test_char_to_boolean_check_db_values(self, model_value, raw_value):
        d = Donut(is_fresh=model_value, is_only_fresh=model_value)
        d.save()
        raw_donut = self._get_raw_donut()

        assert raw_donut.is_fresh == raw_value
        assert raw_donut.is_only_fresh == raw_value



import datetime
import decimal
import uuid

from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.models import Aggregate
from django.db.backends import utils as backend_utils
from django.utils.dateparse import parse_date, parse_datetime, parse_time


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "django_informixdb.compiler"

    def quote_name(self, name):
        return name

    def last_insert_id(self, cursor, table_name, pk_name):
        operation = "SELECT DBINFO('sqlca.sqlerrd1') FROM SYSTABLES WHERE TABID=1"
        cursor.execute(operation)
        row = cursor.fetchone()
        last_identity_val = None
        if row is not None:
            last_identity_val = int(row[0])
        return last_identity_val

    def fulltext_search_sql(self, field_name):
        return "LIKE '%%%s%%'" % field_name

    def lookup_cast(self, lookup_type, internal_type=None):
        if lookup_type in ('iexact', 'icontains', 'istartswith', 'iendswith'):
            return "LOWER(%s)"
        return "%s"

    def check_expression_support(self, expression):
        if isinstance(expression, Aggregate):
            if expression.function in ['STDDEV_POP', 'STDDEV_SAMP']:
                expression.function = 'STDDEV'
            if expression.function in ['VAR_POP', 'VAR_SAMP']:
                expression.function = 'VARIANCE'

    def date_extract_sql(self, lookup_type, field_name):
        sqlmap = {
            'week_day': 'WEEKDAY',
            'month': 'MONTH',
            'day': 'DAY'
        }
        return "%s(%s)" % (sqlmap[lookup_type], field_name)

    def year_lookup_bounds_for_date_field(self, value):
        first = '%s-01-01' % value
        last = '%s-12-31' % value
        return [first, last]

    def start_transaction_sql(self):
        return "BEGIN WORK"

    def end_transaction_sql(self, success=True):
        return "COMMIT WORK"

    def savepoint_create_sql(self, sid):
        return "SAVEPOINT %s" % sid

    def savepoint_commit_sql(self, sid):
        return "RELEASE SAVEPOINT %s" % sid

    def get_db_converters(self, expression):
        converters = super(DatabaseOperations, self).get_db_converters(expression)
        internal_type = expression.output_field.get_internal_type()
        if internal_type == 'BooleanField':
            converters.append(lambda value, *_: True if value == 1 else False)
        elif internal_type == 'NullBooleanField':
            converters.append(lambda value, *_: True if value == 1 else False if value == 0 else None)
        elif internal_type == 'DateTimeField':
            converters.append(self.convert_datetimefield_value)
        elif internal_type == 'DateField':
            converters.append(self.convert_datefield_value)
        elif internal_type == 'TimeField':
            converters.append(self.convert_timefield_value)
        elif internal_type == 'DecimalField':
            converters.append(self.convert_decimalfield_value)
        elif internal_type == 'UUIDField':
            converters.append(self.convert_uuidfield_value)
        return converters

    def convert_decimalfield_value(self, value, expression, connection, *ignore):
        value = backend_utils.format_number(value, expression.output_field.max_digits,
                                            expression.output_field.decimal_places)
        if value is not None:
            return decimal.Decimal(value)

    def convert_datefield_value(self, value, expression, connection, *ignore):
        if value is not None and not isinstance(value, datetime.date):
            value = parse_date(value)
        return value

    def convert_datetimefield_value(self, value, expression, connection, *ignore):
        if value is not None and not isinstance(value, datetime.datetime):
            value = parse_datetime(value)
        return value

    def convert_timefield_value(self, value, expression, connection, *ignore):
        if value is not None and not isinstance(value, datetime.time):
            value = parse_time(value)
        return value

    def convert_uuidfield_value(self, value, expression, connection, *ignore):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def adapt_datefield_value(self, value):
        return value

    def adapt_datetimefield_value(self, value):
        # TODO: fix this, convert to DATETIME YEAR TO FRACTION(5)
        # value is like '2016-05-23 12:26:56.111909+00:00',
        # since informix only support fraction(5),
        # we need remove the last digit for micro-seconds
        return value.strftime('%Y-%m-%d %H:%M:%S.f')[:-1] if value else value

    def adapt_timefield_value(self, value):
        return value

    def sql_flush(self, style, tables, sequences, allow_cascade=False):
        # NB: The generated SQL below is specific to Informix
        sql = ['%s %s %s;' % (
            style.SQL_KEYWORD('DELETE'),
            style.SQL_KEYWORD('FROM'),
            style.SQL_FIELD(self.quote_name(table))
        ) for table in tables]
        return sql

    #def bulk_insert_sql(self, fields, placeholder_rows):
    #    placeholder_rows_sql = (", ".join(row) for row in placeholder_rows)
    #    values_sql = ", ".join("(%s)" % sql for sql in placeholder_rows_sql)
    #    return "VALUES " + values_sql



from django.db.backends.base.introspection import BaseDatabaseIntrospection, FieldInfo, TableInfo

from .datatypes import InformixTypes
from .tableignore import EXCLUDED_TABLES


class DatabaseIntrospection(BaseDatabaseIntrospection):
    # Map type codes to Django Field types.
    data_types_reverse = InformixTypes.field_map()

    def get_table_list(self, cursor):
        cursor.execute('SELECT tabname, tabtype FROM systables')
        # We want to ignore all the internal 'sys' tables in the exclude list
        return [
            TableInfo(x[0], x[1].lower()) for x in cursor.fetchall()
            if x[0] not in EXCLUDED_TABLES
        ]

    def get_table_description(self, cursor, table_name):
        "Returns a description of the table, with the DB-API cursor.description interface."
        query_format = """
            SELECT c.*
            FROM syscolumns c
            JOIN systables t
            ON c.tabid=t.tabid
            WHERE t.tabname='{}'
        """

        cursor.execute(query_format.format(table_name))
        columns = [[c[0], c[3] % 256, None, c[4], c[4], None, 0 if c[3] > 256 else 1, None]
                   for c in cursor.fetchall()]
        items = []
        for column in columns:
            if column[1] in (InformixTypes.SQL_TYPE_NUMERIC.num, InformixTypes.SQL_TYPE_DECIMAL.num):
                column[4] = int(column[3] / 256)
                column[5] = column[3] - column[4] * 256
            # FieldInfo:
            #   name, type_code, display_size, internal_size, precision, scale, null_ok, default
            items.append(FieldInfo(*column))

        return items

    def get_key_columns(self, cursor, table_name):
        key_columns_query = """
            SELECT col1.colname AS column_name,
                   t2.tabname AS referenced_table_name,
                   col2.colname AS referenced_column
            FROM systables t1
            JOIN syscolumns col1 ON t1.tabid = col1.tabid
            JOIN sysindexes idx1 ON idx1.tabid=t1.tabid AND col1.colno = idx1.part1
            JOIN sysconstraints const1 ON idx1.idxname = const1.idxname AND const1.tabid = t1.tabid
            JOIN sysreferences  ref ON ref.constrid = const1.constrid
            JOIN sysconstraints const2 ON ref.primary = const2.constrid
            JOIN sysindexes idx2 ON idx2.idxname = const2.idxname
            JOIN syscolumns col2 ON col2.colno = idx2.part1 AND col2.tabid = idx2.tabid
            JOIN systables t2 ON t2.tabid = idx2.tabid
            WHERE const1.constrtype='R' AND t1.tabname ='{}'
        """
        cursor.execute(key_columns_query.format(table_name))
        return cursor.fetchall()

    def get_indexes(self, cursor, table_name):
        """ This query retrieves each index ON the given table, including the
            first associated field name """
        index_query = """
            SELECT
                sc.colname, idx.idxtype, scs.constrtype
            FROM systables st JOIN sysindexes idx  ON st.tabid = idx.tabid
            JOIN syscolumns sc ON idx.part1=sc.colno AND sc.tabid = st.tabid
            LEFT JOIN sysconstraints scs ON idx.idxname = scs.idxname
            WHERE st.tabname='{}' AND idx.part2 = 0
        """
        cursor.execute(index_query.format(table_name))
        indexes = {}
        for row in cursor.fetchall():
            indexes[row[0]] = {
                'primary_key': True if row[2] == 'P' else False,
                'unique': True if row[1] == 'U' else False
            }
        return indexes

    def _get_col_index(self, cursor, table_name):
        """Private method. Getting Index position of column by its name"""
        col_query = """
            SELECT colno, colname FROM syscolumns sc
            JOIN systables st ON sc.tabid=st.tabid
            WHERE st.tabname='{}'
        """
        cursor.execute(col_query.format(table_name))
        return {row[1]: int(row[0]) - 1 for row in cursor.fetchall()}

    def get_relations(self, cursor, table_name):
        """
        Returns a dictionary of {field_index: (field_index_other_table, other_table)}
        representing all relationships to the given table. Indexes are 0-based.
        """
        relations = {}
        key_columns = self.get_key_columns(cursor, table_name)
        for rel in key_columns:
            row0 = self._get_col_index(cursor, table_name)[rel[0]]
            row1 = self._get_col_index(cursor, rel[1])[rel[2]]
            row2 = rel[1]
            relations[row0] = (row1, row2)
        return relations

    def get_constraints(self, cursor, table_name):
        constraints = {}
        index_query = """
            SELECT idxname, idxtype, indexkeys
            FROM sysindices idx
            JOIN systables st ON idx.tabid = st.tabid
            WHERE st.tabname = '{}'
        """

        # reverse name, AND index here
        all_columns = {v: k for k, v in self._get_col_index(cursor, table_name).items()}
        cursor.execute(index_query.format(table_name))
        for name, idx_type, keys in cursor.fetchall():
            # keys are in the format like "1 [1], 4 [1], 7 [1]",
            # which means including column #1, #4, AND #7
            columns = [all_columns[int(k.strip().split()[0]) - 1] for k in keys.split(',')]
            constraints[name] = {
                'columns': columns,
                'primary_key': len(columns) == 1 and idx_type == 'U',
                'unique': idx_type == 'U',
                'foreign_key': None,
                'check': True,
                'index': idx_type == 'D'
            }
        return constraints



from unittest.mock import Mock, call
from datetime import timedelta

import pyodbc
import pytest
from freezegun import freeze_time

from django_informixdb.base import DatabaseWrapper


CONNECTION_FAILED_ERROR = pyodbc.Error(
    "08004", "Attempt to connect to database server (<dbname>) failed. (-908) (SQLDriverConnect)"
)
READ_ERROR = pyodbc.Error(
    "HY000", "Read error occurred during connection attempt. (-27001) (SQLDriverConnect)"
)
AUTHENTICATION_ERROR = pyodbc.Error(
    "28000",
    "Incorrect password or user is not known on the database server (-951) (SQLDriverConnect)"
)


@pytest.fixture
def db_config():
    return {
        "ENGINE": "django_informixdb",
        "SERVER": "informix",
        "NAME": "sysmaster",
        "USER": "informix",
        "PASSWORD": "in4mix",
        "OPTIONS": {},
        "AUTOCOMMIT": True,
        "CONN_MAX_AGE": None,
        "TIME_ZONE": None,
    }


@pytest.fixture
def mock_autocommit_methods(mocker):
    mocker.patch.object(DatabaseWrapper, "get_autocommit", return_value=True, autospec=True)
    mocker.patch.object(DatabaseWrapper, "set_autocommit", autospec=True)


@pytest.fixture
def mock_is_usable(mocker):
    yield mocker.patch.object(DatabaseWrapper, "is_usable", return_value=True, autospec=True)


@pytest.fixture
def mock_connect(mocker):
    yield mocker.patch("pyodbc.connect", autospec=True)


@pytest.fixture
def mock_connection(mock_connect):
    yield mock_connect.return_value


@pytest.fixture
def mock_sleep(mocker):
    yield mocker.patch("time.sleep", autospec=True)


def test_DatabaseWrapper_connect_successfully_connects(mock_autocommit_methods, db_config):
    db = DatabaseWrapper(db_config)
    assert db.connection is None
    db.connect()
    assert db.connection is not None
    assert db.is_usable() is True


def test_DatabaseWrapper_validate_connection_closes_connections_that_are_not_usable(
    mock_autocommit_methods, mock_is_usable, db_config
):
    mock_is_usable.return_value = False
    db = DatabaseWrapper({
        **db_config,
        "OPTIONS": {"VALIDATE_CONNECTION": True, "VALIDATION_INTERVAL": 0},
    })
    db.connect()
    db.validate_connection()
    assert db.connection is None


def test_DatabaseWrapper_validate_connection_does_not_close_connection_if_not_enabled(
    mock_autocommit_methods, mock_is_usable, db_config
):
    mock_is_usable.return_value = False
    db = DatabaseWrapper({
        **db_config,
        "OPTIONS": {"VALIDATE_CONNECTION": False, "VALIDATION_INTERVAL": 0},
    })
    db.connect()
    db.validate_connection()
    assert db.connection is not None


def test_DatabaseWrapper_validate_connection_handles_closed_connections(
    mock_autocommit_methods, db_config
):
    db = DatabaseWrapper({
        **db_config,
        "OPTIONS": {"VALIDATE_CONNECTION": True, "VALIDATION_INTERVAL": 0},
        "CONN_MAX_AGE": 0,  # this will make close_if_unusable_or_obsolete close the connection
    })
    db.connect()
    db.validate_connection()
    assert db.connection is None


def test_connection_is_not_revalidated_within_validation_interval(
    mock_is_usable, mock_autocommit_methods, db_config
):
    with freeze_time() as frozen_time:
        db = DatabaseWrapper({
            **db_config,
            "OPTIONS": {
                "VALIDATE_CONNECTION": True,
                "VALIDATION_INTERVAL": 10,
            },
        })
        db.connect()

        frozen_time.tick(delta=timedelta(seconds=10))
        db.validate_connection()
        assert mock_is_usable.called is True

        mock_is_usable.reset_mock()

        frozen_time.tick(delta=timedelta(seconds=5))
        db.validate_connection()
        assert mock_is_usable.called is False

        frozen_time.tick(delta=timedelta(seconds=5))
        db.validate_connection()
        assert mock_is_usable.called is True


def test_DatabaseWrapper_is_usable(mocker, mock_autocommit_methods, db_config):
    mocker.patch("pyodbc.connect", autospec=True)
    db = DatabaseWrapper(db_config)
    db.connect()
    assert db.is_usable() is True


def test_DatabaseWrapper_is_usable_returns_false_if_creating_cursor_fails(
    caplog, mocker, mock_connection, mock_autocommit_methods, db_config
):
    mock_connection.cursor.side_effect = pyodbc.Error("", "error message 1")
    db = DatabaseWrapper(db_config)
    db.connect()
    assert db.is_usable() is False
    assert "error message 1" in caplog.text


def test_DatabaseWrapper_is_usable_returns_false_if_execute_raises_error(
    caplog, mocker, mock_connection, mock_autocommit_methods, db_config
):
    mock_connection.cursor.return_value.execute.side_effect = pyodbc.Error("", "error message 1")
    db = DatabaseWrapper(db_config)
    db.connect()
    assert db.is_usable() is False
    assert "error message 1" in caplog.text


def test_DatabaseWrapper_is_usable_returns_false_if_closing_cursor_raises_error(
    caplog, mocker, mock_connection, mock_autocommit_methods, db_config
):
    mock_connection.cursor.return_value.close.side_effect = pyodbc.Error("", "error message 1")
    db = DatabaseWrapper(db_config)
    db.connect()
    assert db.is_usable() is False
    assert "error message 1" in caplog.text


def test_DatabaseWrapper_get_new_connection_calls_pyodbc_connect(mock_connect, db_config):
    db = DatabaseWrapper({
        "SERVER": "server1",
        "NAME": "db1",
        "USER": "user1",
        "PASSWORD": "password1",
        "OPTIONS": {"CONN_TIMEOUT": 120},
    })
    db.get_new_connection(db.get_connection_params())
    assert mock_connect.called is True
    connection_string = mock_connect.call_args[0][0]
    parts = connection_string.split(";")
    assert "Server=server1" in parts
    assert "Database=db1" in parts
    assert "Uid=user1" in parts
    assert "Pwd=password1" in parts
    assert mock_connect.call_args[1]['timeout'] == 120


def test_get_new_connection_doesnt_retry_by_default(mock_connect, db_config):
    mock_connect.side_effect = READ_ERROR
    db = DatabaseWrapper(db_config)
    with pytest.raises(pyodbc.Error):
        db.get_new_connection(db.get_connection_params())
    assert mock_connect.call_count == 1


def test_get_new_connection_retries_up_to_MAX_ATTEMPTS(mock_sleep, mock_connect, db_config):
    mock_connect.side_effect = READ_ERROR
    db = DatabaseWrapper({**db_config, "CONNECTION_RETRY": {"MAX_ATTEMPTS": 3}})
    with pytest.raises(pyodbc.Error):
        db.get_new_connection(db.get_connection_params())
    assert mock_connect.call_count == 3
    assert mock_sleep.call_count == 2


@pytest.mark.parametrize("error, expected_retries", [
    (CONNECTION_FAILED_ERROR, 1),
    (READ_ERROR, 1),
    (AUTHENTICATION_ERROR, 0),
    (pyodbc.Error("", "-27001"), 0),  # the error code must be surrounded by round brackets
])
def test_only_retries_certain_errors(mock_sleep, mock_connect, db_config, error, expected_retries):
    mock_connect.side_effect = error
    db = DatabaseWrapper({**db_config, "CONNECTION_RETRY": {"MAX_ATTEMPTS": 2}})
    with pytest.raises(pyodbc.Error):
        db.get_new_connection(db.get_connection_params())
    assert mock_connect.call_count == 1 + expected_retries  # attempts = 1 + retries


@pytest.mark.parametrize("error, expected_retries", [
    (CONNECTION_FAILED_ERROR, 0),
    (pyodbc.Error("", "(-12345)"), 1),
])
def test_errors_to_retry_can_be_overridden(mock_sleep, mock_connect, db_config, error, expected_retries):
    mock_connect.side_effect = error
    db = DatabaseWrapper({
        **db_config,
        "CONNECTION_RETRY": {
            "MAX_ATTEMPTS": 2,
            "ERRORS": ["-12345"],
        },
    })
    with pytest.raises(pyodbc.Error):
        db.get_new_connection(db.get_connection_params())
    assert mock_connect.call_count == 1 + expected_retries


def test_get_new_connection_breaks_early_if_connection_succeeds(
    mock_sleep, mock_connect, db_config
):
    mock_connect.side_effect = [READ_ERROR, Mock()]
    db = DatabaseWrapper({**db_config, "CONNECTION_RETRY": {"MAX_ATTEMPTS": 3}})
    db.get_new_connection(db.get_connection_params())
    assert mock_connect.call_count == 2
    assert mock_sleep.call_count == 1


def test_sleeps_between_connection_attempts(mocker, mock_sleep, mock_connect, db_config):
    mock_uniform = mocker.patch(
        "random.uniform",
        autospec=True,
        side_effect=[1000, 2000, 3000, 4000, 5000],
    )
    mock_connect.side_effect = READ_ERROR
    db = DatabaseWrapper({
        **db_config,
        "CONNECTION_RETRY": {
            "MAX_ATTEMPTS": 6,
            "WAIT_MIN": 15,
            "WAIT_MAX": 100,
            "WAIT_MULTIPLIER": 10,
            "WAIT_EXP_BASE": 2,
        },
    })
    with pytest.raises(pyodbc.Error):
        db.get_new_connection(db.get_connection_params())
    assert mock_uniform.call_args_list == [
        call(15, 15),
        call(15, 20),
        call(15, 40),
        call(15, 80),
        call(15, 100),
    ]
    assert mock_sleep.call_args_list == [call(1), call(2), call(3), call(4), call(5)]



import pytest
from django.conf import settings


def pytest_configure():
    settings.configure(
        ROOT_URLCONF="test.urls",
        MIDDLEWARE=(),
        INSTALLED_APPS=(
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "test.datatypes",
        ),
        DATABASES={
            "default": {
                "ENGINE": "django_informixdb",
                "SERVER": "informix",
                "NAME": "adapter",
                "USER": "informix",
                "PASSWORD": "in4mix",
            },
        },
    )


@pytest.fixture(autouse=True)
def configure_caplog(caplog):
    caplog.set_level("INFO")



from django.db.models.sql import compiler


class SQLCompiler(compiler.SQLCompiler):
    def as_sql(self, with_limits=True, with_col_aliases=False):
        raw_sql, fields = super(SQLCompiler, self).as_sql(False, with_col_aliases)

        # special dialect to return first n rows
        if with_limits:
            if self.query.high_mark is not None:
                _select = "SELECT"
                _first = self.query.high_mark
                if self.query.low_mark:
                    _select += " SKIP %s" % self.query.low_mark
                    _first -= self.query.low_mark
                _select += " FIRST %s" % _first
                raw_sql = raw_sql.replace("SELECT", _select, 1)

        return raw_sql.replace(r'%s', '?'), fields


def _list2tuple(arg):
    return tuple(arg) if isinstance(arg, list) else arg


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):
    def as_sql(self):
        result = super(SQLInsertCompiler, self).as_sql()
        return [(ret[0].replace(r'%s', '?'), _list2tuple(ret[1])) for ret in result]


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    def as_sql(self):
        result = super(SQLAggregateCompiler, self).as_sql()
        return result[0].replace(r'%s', '?'), result[1]


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    def as_sql(self):
        result = super(SQLDeleteCompiler, self).as_sql()
        return result[0].replace(r'%s', '?'), result[1]


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    def as_sql(self):
        result = super(SQLUpdateCompiler, self).as_sql()
        return result[0].replace(r'%s', '?'), result[1]



import django.db


def test_request_triggers_validate_connection(mocker, db, client):
    mock_validate_connection = mocker.patch.object(
        django.db.connection, "validate_connection", autospec=True
    )
    client.get("")
    assert mock_validate_connection.called is True
